;;; DO NOT MODIFY THIS FILE
(if (featurep 'crisp-autoloads) (error "Already loaded"))

;;;### (autoloads nil "_pkg" "crisp/_pkg.el")

(package-provide 'crisp :version 1.13 :type 'regular)

;;;***

(provide 'crisp-autoloads)
